/* riffwav.c.  R.W Dobson 8.99 */
#include "dpwelib.h"
#include "byteswap.h"	/* not using very much of this */
#include "riffwav.h"
#include <memory.h>
#include <math.h>
#include <time.h>
#include	"IEEE80.h"
/* RWD update log:
 * 24.8.99 completed wavTell
 *         corrected error return values for wavWrite~ funcs
 * 29.8.99 completed wavWriteFloatFrames, changed outbuf to floats
 */


#ifndef DBGFPRINTF
# ifdef _DEBUG
# define DBGFPRINTF(a) fprintf a
# else
# define DBGFPRINTF(a)
# endif
#endif

#ifndef BITS_PER_BYTE
#define BITS_PER_BYTE (8)
#endif
/* probably no good for 64bit platforms */
#define REVDWBYTES(t)	( (((t)&0xff) << 24) | (((t)&0xff00) << 8) | (((t)&0xff0000) >> 8) | (((t)>>24) & 0xff) )
#define REVWBYTES(t)	( (((t)&0xff) << 8) | (((t)>>8) &0xff) )
#define TAG(a,b,c,d)	( ((a)<<24) | ((b)<<16) | ((c)<<8) | (d) )
#define MAX_16BIT  (32768.0)
#define MAX_32BIT  (2147483648.0)
#define AIFC_VERSION_1 (0xA2805140)
//pstring for AIFC	- includes the pad byte
static const char aifc_floatstring[10] = { 0x08,'F','l','o','a','t',0x20,'3','2',0x00};
static const char aifc_notcompressed[16] = {0x0e,'n','o','t',0x20,'c','o','m','p','r','e','s','s','e','d',0x00};
 
static void fmtSwapBytes(RIFFWAV_STRUCT *wavs)
{
	WAVEFORMAT *pfmt		=  &(wavs->fmt);
	pfmt->wFormatTag		= REVWBYTES(pfmt->wFormatTag);
	pfmt->nChannels			= REVWBYTES(pfmt->nChannels);
	pfmt->nSamplesPerSec	= REVDWBYTES(pfmt->nSamplesPerSec);
	pfmt->nAvgBytesPerSec	= REVDWBYTES(pfmt->nAvgBytesPerSec);
	pfmt->nBlockAlign		= REVWBYTES(pfmt->nBlockAlign);
	pfmt->wBitsPerSample	= REVWBYTES(pfmt->wBitsPerSample);

}
/*RWD: don't think I need this.... */
static void ieee_swap_bytes(unsigned char ieeechar[])
{
	unsigned char tmp[10], *ptmp,*pieee;
	int i;
	ptmp = tmp+10;
	pieee = ieeechar;

	for(i=10;i;--i)
		*(--ptmp) = *pieee++;
	pieee = ieeechar;
	for(i=10;i;--i)
		*pieee++ = *ptmp++;
}

#if defined _WIN32 && defined _MSC_VER
__inline long round(double fval)
{
	int result;
	_asm{
		fld		fval
		fistp	result
		mov		eax,result
	}
	return result;
}

#else
long round(double val)
{
	long k;
	k = (long)(fabs(val)+0.5);
	if(val < 0.0)
		k = -k;
	return k;
}
#endif

RIFFWAV_STRUCT *wavNew(void)
{
	RIFFWAV_STRUCT *wavs;
	/* I avoid TMMALLOC as I don't like it very much!*/
	wavs = (RIFFWAV_STRUCT *) malloc(sizeof(RIFFWAV_STRUCT));
	if(wavs==NULL)
		return wavs;
		
	wavs->file				= NULL;
	wavs->filename			= NULL;
	wavs->nFrames			= 0;
	wavs->filesize			= 0;				/* used for writing only */
	wavs->curframepos		= 0;				
	wavs->isRead			= 1;				/* OK. who knows?    */
	wavs->riff_format		= RIFF_STDWAVE;		/* almost certainly! */
	wavs->isSeekable		= 1;
	wavs->clip_floats		= 1;
	wavs->rescale			= 0;
	wavs->rescale_fac		= 1.0f;
	wavs->bytemode			= BYWDREV;			/* well, on intel....*/
	wavs->samptype			= SAMP_16;		/* reasonable...     */
	wavs->dataoffset		= 0;
	wavs->fmtoffset			= 0;
	wavs->peakoffset		= 0;
	/*we might as well init WAVEFORMAT to something */
	wavs->fmt.wFormatTag	= WAVE_FORMAT_PCM; 
    wavs->fmt.nChannels		= 1; 
    wavs->fmt.nSamplesPerSec  = 44100; 
    wavs->fmt.nAvgBytesPerSec = wavs->fmt.nSamplesPerSec * wavs->fmt.nChannels * sizeof(short); 
    wavs->fmt.nBlockAlign	  = wavs->fmt.nChannels * sizeof(short); 
    wavs->fmt.wBitsPerSample  = sizeof(short) * BITS_PER_BYTE;
	wavs->pPeaks			= NULL;
	wavs->peaktime			= 0;
	return wavs;
}

static int wavUpdate(RIFFWAV_STRUCT *wavs)
{
	long riffsize,datasize;
	
	if(wavs==NULL || wavs->file== NULL)
		return RIFFWAV_E_BADARG;

	if(wavs->dataoffset==0)
		return RIFFWAV_E_BADARG;
	
	if((fseek(wavs->file,sizeof(long),SEEK_SET))==0) {
		riffsize = wavs->filesize-8;
		if(wavs->bytemode ==INORDER)
			riffsize = REVDWBYTES(riffsize);
		if(fwrite((char *) &riffsize,sizeof(long),1,wavs->file) != 1)
			return RIFFWAV_E_CANT_WRITE;
	}

	if(wavs->pPeaks){
		if(wavs->peakoffset==0)
			return RIFFWAV_E_BADARG;
		
		//do byterev if necessary...
		if((fseek(wavs->file,wavs->peakoffset,SEEK_SET))==0){
			//set current time
			DWORD *pblock;
			int i;
			time_t now = time(0);
			if(wavs->bytemode ==INORDER){
				now = REVDWBYTES(now);
				pblock = (DWORD *) (wavs->pPeaks);
				for(i=0;i < wavs->fmt.nChannels * 2; i++)
					pblock[i] = REVDWBYTES(pblock[i]);
			}
			if((fwrite((char*)&now,sizeof(DWORD),1,wavs->file)) != 1)
				return RIFFWAV_E_CANT_WRITE;

			if((fwrite((char *) (wavs->pPeaks),sizeof(WAV_CHPEAK),wavs->fmt.nChannels,wavs->file))
				!= wavs->fmt.nChannels )
				return RIFFWAV_E_CANT_WRITE;
		}
	}

	if((fseek(wavs->file,wavs->dataoffset-sizeof(long),SEEK_SET))==0) {			
		datasize = wavs->nFrames * wavs->fmt.nBlockAlign;
		if(wavs->bytemode ==INORDER)
			datasize = REVDWBYTES(datasize);
		if(fwrite((char *) & datasize,sizeof(long),1,wavs->file) != 1)
			return RIFFWAV_E_CANT_WRITE;	
	}
	if(fseek(wavs->file,0,SEEK_END)){
		DBGFPRINTF((stderr,"wavUpdate: error reseeking to end of file\n"));
	}

	return RIFFWAV_E_NOERROR;
}

static int aiffUpdate(RIFFWAV_STRUCT *wavs)
{
	long aiffsize,datasize,frames,filesize;
	unsigned char pad = 0x00;
	if(wavs==NULL || wavs->file== NULL)
		return RIFFWAV_E_BADARG;

	if(wavs->dataoffset==0)
		return RIFFWAV_E_BADARG;
	
	if((fseek(wavs->file,sizeof(long),SEEK_SET))==0) {
		aiffsize = wavs->filesize-8;
		if(wavs->bytemode !=INORDER)
			aiffsize = REVDWBYTES(aiffsize);
		if(fwrite((char *) &aiffsize,sizeof(long),1,wavs->file) != 1)
			return RIFFWAV_E_CANT_WRITE;
	}
	if((fseek(wavs->file,wavs->fmtoffset + sizeof(WORD),SEEK_SET))==0) {
		frames = wavs->nFrames;		
		if(wavs->bytemode !=INORDER)
			frames = REVDWBYTES(frames);
		if(fwrite((char *) &frames,sizeof(long),1,wavs->file) != 1)
			return RIFFWAV_E_CANT_WRITE;
	}
	if(wavs->pPeaks){
		if(wavs->peakoffset==0)
			return RIFFWAV_E_BADARG;
		
		//do byterev if necessary...
		if((fseek(wavs->file,wavs->peakoffset,SEEK_SET))==0){
			//set current time
			DWORD *pblock;
			int i;
			time_t now = time(0);
			if(wavs->bytemode !=INORDER){
				now = REVDWBYTES(now);
				pblock = (DWORD *) (wavs->pPeaks);
				for(i=0;i < wavs->fmt.nChannels * 2; i++)
					pblock[i] = REVDWBYTES(pblock[i]);
			}
			if((fwrite((char*)&now,sizeof(DWORD),1,wavs->file)) != 1)
				return RIFFWAV_E_CANT_WRITE;

			if((fwrite((char *) (wavs->pPeaks),sizeof(WAV_CHPEAK),wavs->fmt.nChannels,wavs->file))
				!= wavs->fmt.nChannels )
				return RIFFWAV_E_CANT_WRITE;
		}
	}

	if((fseek(wavs->file,wavs->dataoffset- (3 *  sizeof(long)),SEEK_SET))==0) {			
		datasize = wavs->nFrames * wavs->fmt.nBlockAlign;
		datasize += 2* sizeof(DWORD);	/* add offset and blocksize fields */
		if(wavs->bytemode !=INORDER)
			datasize = REVDWBYTES(datasize);
		if(fwrite((char *) & datasize,sizeof(long),1,wavs->file) != 1)
			return RIFFWAV_E_CANT_WRITE;	
	}

	/* datachunk needs added pad byte if odd, not included in saved chunksize*/
	if((fseek(wavs->file,wavs->dataoffset+datasize,SEEK_SET))==0){
		return RIFFWAV_E_CANT_SEEK;
	}
	filesize = ftell(wavs->file);
	if(filesize % 2)
		if(fwrite(&pad,sizeof(unsigned char),1,wavs->file) != 1)
			return RIFFWAV_E_CANT_WRITE;
#ifdef NOTDEF
	if(fseek(wavs->file,0,SEEK_END)){
		DBGFPRINTF((stderr,"aiffUpdate: error reseeking to end of file\n"));
	}
#endif
	return RIFFWAV_E_NOERROR;
}



static int wavDoWrite(RIFFWAV_STRUCT *wavs, char *buf, DWORD nBytes)
{
	
	DWORD written = 0;
	if(wavs==NULL || buf==NULL)
		return RIFFWAV_E_BADARG;

	if(wavs->file==NULL)
		return RIFFWAV_E_CANT_WRITE;

	if((written = fwrite(buf,sizeof(char),nBytes,wavs->file)) != nBytes) {
		DBGFPRINTF((stderr, "wavDoWrite: wanted %ld got %ld.\n",
                    nBytes, written));
        return RIFFWAV_E_CANT_WRITE;
    }
	
	wavs->filesize += written;		//be more clever?

	return RIFFWAV_E_NOERROR;

}

static int wavDoRead(RIFFWAV_STRUCT *wavs, char *buf, DWORD nBytes)
{
	
	DWORD got = 0;
	if(wavs==NULL || buf==NULL)
		return RIFFWAV_E_BADARG;

	if(wavs->file==NULL)
		return RIFFWAV_E_CANT_READ;

	if((got = fread(buf,sizeof(char),nBytes,wavs->file)) != nBytes) {
		DBGFPRINTF((stderr, "wavDoWrite: wanted %ld got %ld.\n",
                    nBytes, got));
        return RIFFWAV_E_CANT_READ;
    }

	return RIFFWAV_E_NOERROR;

}

//test all this, then do byterev stuff
static int wavWriteHeader(RIFFWAV_STRUCT *wavs)
{
	long tag,size;
	WORD cbSize = 0;
	WAVEFORMAT *pfmt;
	WAV_CHPEAK *peaks;

	if(wavs==NULL || wavs->file==NULL)
		return RIFFWAV_E_BADARG;

	if(wavs->riff_format!= RIFF_STDWAVE)		/*for now*/
		return RIFFWAV_E_BADARG;

	if(wavs->nFrames > 0 || wavs->isRead)
		return RIFFWAV_E_BADARG;

	if(wavs->fmt.nChannels == 0)
		return RIFFWAV_E_BADARG;

	//clear pPeaks array
	if(wavs->pPeaks)
		memset((char *)wavs->pPeaks,0,sizeof(WAV_CHPEAK) * wavs->fmt.nChannels);


	tag = TAG('R','I','F','F');
	size = 0;
	if(wavs->bytemode==INORDER)
		size = REVDWBYTES(size);
	if(wavs->bytemode!=INORDER)
		tag = REVDWBYTES(tag);
	

	if(wavDoWrite(wavs,(char *)&tag,sizeof(long))
		|| wavDoWrite(wavs,(char *)&size,sizeof(long)))
		return RIFFWAV_E_CANT_WRITE;

	tag = TAG('W','A','V','E');
	if(wavs->bytemode!=INORDER)
		tag = REVDWBYTES(tag);
	if(wavDoWrite(wavs,(char *)&tag,sizeof(long)))
		return RIFFWAV_E_CANT_WRITE;

	pfmt = &(wavs->fmt);

	tag = TAG('f','m','t',' ');	
	size = sizeof(WAVEFORMAT);
	if(wavs->samptype==SAMP_IEEE_FLOAT)
		size += sizeof(WORD);		  /* for cbSize: WAVEOFRMATEX */
	if(wavs->bytemode==INORDER){
		size = REVDWBYTES(size);
		fmtSwapBytes(wavs);
	}
	if(wavs->bytemode!=INORDER)
		tag = REVDWBYTES(tag);
	if(wavDoWrite(wavs,(char *)&tag,sizeof(long))
		|| wavDoWrite(wavs,(char *)&size,sizeof(long)))		
		return RIFFWAV_E_CANT_WRITE;

	wavs->fmtoffset = ftell(wavs->file);
	
	if(wavDoWrite(wavs,(char *)pfmt,sizeof(WAVEFORMAT)))
		return RIFFWAV_E_CANT_WRITE;
	/*add cbSize if floatsams */
	if(wavs->samptype==SAMP_IEEE_FLOAT)
		if(wavDoWrite(wavs,(char *)&cbSize,sizeof(WORD)))
			return RIFFWAV_E_CANT_WRITE;
	if(wavs->pPeaks){
		DWORD version = 1, now = 0;
		peaks = wavs->pPeaks;
		
		tag = TAG('P','E','A','K');		
		size = 2 * sizeof(DWORD) + sizeof(WAV_CHPEAK) * pfmt->nChannels;
		if(wavs->bytemode==INORDER){
			size = REVDWBYTES(size);
			version  = REVDWBYTES(version);
		}
		if(wavs->bytemode!=INORDER)
			tag = REVDWBYTES(tag);


		if(wavDoWrite(wavs,(char *)&tag,sizeof(long))
			|| wavDoWrite(wavs,(char *)&size,sizeof(long))
			|| wavDoWrite(wavs,(char *)&version,sizeof(long)))
			return RIFFWAV_E_CANT_WRITE;
		wavs->peakoffset = ftell(wavs->file);  //we need to update time

		if(wavDoWrite(wavs,(char *) &now,sizeof(long))
			|| wavDoWrite(wavs,(char *) peaks, sizeof(WAV_CHPEAK) * wavs->fmt.nChannels))
			return RIFFWAV_E_CANT_WRITE;
	}
	
	tag = TAG('d','a','t','a');	
	size = 0;	
	if(wavs->bytemode!=INORDER)
		tag = REVDWBYTES(tag);

	if(wavDoWrite(wavs,(char *)&tag,sizeof(long))
		|| wavDoWrite(wavs,(char *)&size,sizeof(long)))		
		return RIFFWAV_E_CANT_WRITE;
	wavs->dataoffset = ftell(wavs->file);
	wavs->filesize = wavs->dataoffset;
	return RIFFWAV_E_NOERROR;
}

static int aiffWriteHeader(RIFFWAV_STRUCT *wavs)
{
	long tag,size;		
	WAV_CHPEAK *peaks;	
	DWORD dwData,offset,blocksize;
	unsigned char ieee[10];
	WORD wData;

	if(wavs==NULL || wavs->file==NULL)
		return RIFFWAV_E_BADARG;

	if(wavs->riff_format!=FORM_AIFF)		
		return RIFFWAV_E_BADARG;

	if(wavs->nFrames > 0 || wavs->isRead)
		return RIFFWAV_E_BADARG;

	if(wavs->fmt.nChannels == 0)
		return RIFFWAV_E_BADARG;

	//clear pPeaks array
	if(wavs->pPeaks)
		memset((char *)wavs->pPeaks,0,sizeof(WAV_CHPEAK) * wavs->fmt.nChannels);

	tag = TAG('F','O','R','M');
	size = 0;

	if(wavs->bytemode!=INORDER) {
		size = REVDWBYTES(size);	
		tag = REVDWBYTES(tag);
	}
	if(wavDoWrite(wavs,(char *)&tag,sizeof(long))
		|| wavDoWrite(wavs,(char *)&size,sizeof(long)))
		return RIFFWAV_E_CANT_WRITE;
	tag = TAG('A','I','F','F');
	if(wavs->bytemode!=INORDER)
		tag = REVDWBYTES(tag);
	if(wavDoWrite(wavs,(char *)&tag,sizeof(long)))
		return RIFFWAV_E_CANT_WRITE;

	tag = TAG('C','O','M','M');
	size = 18;
	if(wavs->bytemode!=INORDER){
		size = REVDWBYTES(size);	
		tag = REVDWBYTES(tag);
	}
	if(wavDoWrite(wavs,(char *)&tag,sizeof(long))
		|| wavDoWrite(wavs,(char *)&size,sizeof(long)))
		return RIFFWAV_E_CANT_WRITE;

	wavs->fmtoffset = ftell(wavs->file);

	wData = wavs->fmt.nChannels;
	if(wavs->bytemode!=INORDER)
		wData = REVWBYTES(wData);
	if(wavDoWrite(wavs,(char *)&wData,sizeof(WORD)))
		return RIFFWAV_E_CANT_WRITE;

	dwData = 0;			/* nFrames */
	if(wavDoWrite(wavs,(char *)&dwData,sizeof(DWORD)))
		return RIFFWAV_E_CANT_WRITE;

	wData = wavs->fmt.wBitsPerSample;
	if(wavs->bytemode!=INORDER)
		wData = REVWBYTES(wData);
	if(wavDoWrite(wavs,(char *)&wData,sizeof(WORD)))
		return RIFFWAV_E_CANT_WRITE;
	double_to_ieee_80((double)wavs->fmt.nSamplesPerSec,ieee);
	//if(wavs->bytemode==INORDER)
	//	ieee_swap_bytes(ieee);
	if(wavDoWrite(wavs,ieee,10))
		return RIFFWAV_E_CANT_WRITE;
	if(wavs->pPeaks){
		DWORD version = 1, now = 0;
		peaks = wavs->pPeaks;
		
		tag = TAG('P','E','A','K');		
		size = 2 * sizeof(DWORD) + sizeof(WAV_CHPEAK) * wavs->fmt.nChannels;
		if(wavs->bytemode!=INORDER){
			size = REVDWBYTES(size);
			version  = REVDWBYTES(version);
			tag = REVDWBYTES(tag);
		}
		if(wavDoWrite(wavs,(char *)&tag,sizeof(long))
			|| wavDoWrite(wavs,(char *)&size,sizeof(long))
			|| wavDoWrite(wavs,(char *)&version,sizeof(long)))
			return RIFFWAV_E_CANT_WRITE;
		wavs->peakoffset = ftell(wavs->file);  //we need to update time

		if(wavDoWrite(wavs,(char *) &now,sizeof(long))
			|| wavDoWrite(wavs,(char *) peaks, sizeof(WAV_CHPEAK) * wavs->fmt.nChannels))
			return RIFFWAV_E_CANT_WRITE;
	}

	tag = TAG('S','S','N','D');
	size = offset = blocksize = 0;
	if(wavs->bytemode!=INORDER)
			tag = REVDWBYTES(tag);
	if(wavDoWrite(wavs,(char *)&tag,sizeof(long))
			|| wavDoWrite(wavs,(char *)&size,sizeof(long))
			|| wavDoWrite(wavs,(char *)&offset,sizeof(long))
			|| wavDoWrite(wavs,(char *)&blocksize,sizeof(long)))
			return RIFFWAV_E_CANT_WRITE;
	wavs->dataoffset = ftell(wavs->file);
	wavs->filesize = wavs->dataoffset;
	return RIFFWAV_E_NOERROR;
}


static int aifcWriteHeader(RIFFWAV_STRUCT *wavs)
{
	long tag,size;		
	WAV_CHPEAK *peaks;	
	DWORD dwData,offset,blocksize,aifcver = AIFC_VERSION_1,ID_compression;
	//assume 32bit floats, but we may be asked to use aifc for integer formats too
	char *str_compressed = (char *) aifc_floatstring;
	long pstring_size = 10;
	unsigned char ieee[10];
	WORD wData;

	if(wavs==NULL || wavs->file==NULL)
		return RIFFWAV_E_BADARG;

	if(wavs->riff_format!=FORM_AIFC)		
		return RIFFWAV_E_BADARG;

	if(wavs->nFrames > 0 || wavs->isRead)
		return RIFFWAV_E_BADARG;

	if(wavs->fmt.nChannels == 0)
		return RIFFWAV_E_BADARG;

	if(wavs->samptype==SAMP_IEEE_FLOAT)
		ID_compression = TAG('F','L','3','2');
	else {
		ID_compression = TAG('N','O','N','E');
		pstring_size = 16;
		str_compressed = (char *) aifc_notcompressed;
	}
	//clear pPeaks array
	if(wavs->pPeaks)
		memset((char *)wavs->pPeaks,0,sizeof(WAV_CHPEAK) * wavs->fmt.nChannels);

	tag = TAG('F','O','R','M');
	size = 0;

	if(wavs->bytemode!=INORDER) {
		size = REVDWBYTES(size);	
		tag = REVDWBYTES(tag);
	}
	if(wavDoWrite(wavs,(char *)&tag,sizeof(long))
		|| wavDoWrite(wavs,(char *)&size,sizeof(long)))
		return RIFFWAV_E_CANT_WRITE;
	tag = TAG('A','I','F','C');
	if(wavs->bytemode!=INORDER)
		tag = REVDWBYTES(tag);
	if(wavDoWrite(wavs,(char *)&tag,sizeof(long)))
		return RIFFWAV_E_CANT_WRITE;

	tag = TAG('F','V','E','R');
	size = sizeof(DWORD);
	if(wavs->bytemode!=INORDER){
		size = REVDWBYTES(size);	
		tag = REVDWBYTES(tag);
		aifcver = REVDWBYTES(aifcver);
	}
	if(wavDoWrite(wavs,(char *)&tag,sizeof(DWORD))
		|| wavDoWrite(wavs,(char *)&size,sizeof(DWORD))
		|| wavDoWrite(wavs,(char *)&aifcver,sizeof(DWORD)))
		return RIFFWAV_E_CANT_WRITE;


	tag = TAG('C','O','M','M');
	size = 22 + pstring_size;
	if(wavs->bytemode!=INORDER){
		size = REVDWBYTES(size);	
		tag = REVDWBYTES(tag);
	}
	if(wavDoWrite(wavs,(char *)&tag,sizeof(long))
		|| wavDoWrite(wavs,(char *)&size,sizeof(long)))
		return RIFFWAV_E_CANT_WRITE;

	wavs->fmtoffset = ftell(wavs->file);

	wData = wavs->fmt.nChannels;
	if(wavs->bytemode!=INORDER)
		wData = REVWBYTES(wData);
	if(wavDoWrite(wavs,(char *)&wData,sizeof(WORD)))
		return RIFFWAV_E_CANT_WRITE;

	dwData = 0;			/* nFrames */
	if(wavDoWrite(wavs,(char *)&dwData,sizeof(DWORD)))
		return RIFFWAV_E_CANT_WRITE;

	wData = wavs->fmt.wBitsPerSample;
	if(wavs->bytemode!=INORDER)
		wData = REVWBYTES(wData);
	if(wavDoWrite(wavs,(char *)&wData,sizeof(WORD)))
		return RIFFWAV_E_CANT_WRITE;
	double_to_ieee_80((double)wavs->fmt.nSamplesPerSec,ieee);
	//if(wavs->bytemode==INORDER)
	//	ieee_swap_bytes(ieee);
	if(wavDoWrite(wavs,ieee,10))
		return RIFFWAV_E_CANT_WRITE;
	/*AIFC bits */
	if(wavs->bytemode!=INORDER)
		ID_compression = REVDWBYTES(ID_compression);
	if(wavDoWrite(wavs,(char *)&ID_compression,sizeof(DWORD)))
		return RIFFWAV_E_CANT_WRITE;
	if(wavDoWrite(wavs,str_compressed,pstring_size))
		return RIFFWAV_E_CANT_WRITE;


	if(wavs->pPeaks){
		DWORD version = 1, now = 0;
		peaks = wavs->pPeaks;
		
		tag = TAG('P','E','A','K');		
		size = 2 * sizeof(DWORD) + sizeof(WAV_CHPEAK) * wavs->fmt.nChannels;
		if(wavs->bytemode!=INORDER){
			size = REVDWBYTES(size);
			version  = REVDWBYTES(version);
			tag = REVDWBYTES(tag);
		}
		if(wavDoWrite(wavs,(char *)&tag,sizeof(long))
			|| wavDoWrite(wavs,(char *)&size,sizeof(long))
			|| wavDoWrite(wavs,(char *)&version,sizeof(long)))
			return RIFFWAV_E_CANT_WRITE;
		wavs->peakoffset = ftell(wavs->file);  //we need to update time

		if(wavDoWrite(wavs,(char *) &now,sizeof(long))
			|| wavDoWrite(wavs,(char *) peaks, sizeof(WAV_CHPEAK) * wavs->fmt.nChannels))
			return RIFFWAV_E_CANT_WRITE;
	}

	tag = TAG('S','S','N','D');
	size = offset = blocksize = 0;
	if(wavs->bytemode!=INORDER)
			tag = REVDWBYTES(tag);
	if(wavDoWrite(wavs,(char *)&tag,sizeof(long))
			|| wavDoWrite(wavs,(char *)&size,sizeof(long))
			|| wavDoWrite(wavs,(char *)&offset,sizeof(long))
			|| wavDoWrite(wavs,(char *)&blocksize,sizeof(long)))
			return RIFFWAV_E_CANT_WRITE;
	wavs->dataoffset = ftell(wavs->file);
	wavs->filesize = wavs->dataoffset;
	return RIFFWAV_E_NOERROR;
}


void wavFree(RIFFWAV_STRUCT *wavs)
{
	/* nothing else to do! */
	if(wavs->filename)
		TMFREE(wavs->filename,"wavFree");
	if(wavs->pPeaks)
		TMFREE(wavs->pPeaks,"wavFree");
	TMFREE(wavs,"wavFree");
}

/* we expect full format info to be set in wavs */
int wavOpenWrite(RIFFWAV_STRUCT *wavs, const char *path,sf_stype stype,int clip_floats)
{		
	
	if(wavs == NULL || path==NULL)
		return RIFFWAV_E_BADARG;
	if(wavs->file || wavs->pPeaks)
		return RIFFWAV_E_CANT_OPEN;
	wavs->clip_floats = clip_floats;
	wavs->bytemode = HostByteMode();
	if(stype==FMT_UNKNOWN)
		return RIFFWAV_E_BADARG;
	if(wavs->samptype == SAMP_UNKNOWN)
		return RIFFWAV_E_BADARG;

	wavs->filename = (char *) malloc(strlen(path)+1);
	if(wavs->filename==NULL) {
		DBGFPRINTF((stderr, "wavOpenWrite: no memory for filename\n"));
		return RIFFWAV_E_NOMEM;
	}
	wavs->pPeaks = (WAV_CHPEAK *) malloc(sizeof(WAV_CHPEAK) * wavs->fmt.nChannels);
	if(wavs->pPeaks==NULL){
		DBGFPRINTF((stderr, "wavOpenWrite: no memory for peak data\n"));
		return RIFFWAV_E_NOMEM;
	}
	if((wavs->file = fopen(path,"wb+"))  == NULL) {
		DBGFPRINTF((stderr, "wavOpenWrite: cannot create '%s'\n", path));
        return RIFFWAV_E_CANT_OPEN;
	}
	
    strcpy(wavs->filename, path);
    wavs->isRead = FALSE;    	
	wavs->filesize = 0;
	wavs->nFrames = 0;

	/* force aif f/p data to go to aifc format */
	if(wavs->samptype==SAMP_IEEE_FLOAT && stype==FORM_AIFF){
		DBGFPRINTF((stderr, "Warning: writing floating point data in AIFC format\n"));
		stype= FORM_AIFC;
	}
	
	switch(stype){
	case(RIFF_STDWAVE):
		wavs->riff_format = stype;
		return wavWriteHeader(wavs);
	case(FORM_AIFF):
		wavs->riff_format = stype;
		return aiffWriteHeader(wavs);
	case(FORM_AIFC):
		wavs->riff_format = stype;
		return aifcWriteHeader(wavs);
	default:
		return RIFFWAV_E_UNSUPPORTED;
	}
}
	

int wavClose(RIFFWAV_STRUCT *wavs)
{
	int rc;
	
	if(wavs==NULL || wavs->file==NULL)
		return RIFFWAV_E_BADARG;

	
	switch(wavs->riff_format){
	case(RIFF_STDWAVE):
		rc = wavUpdate(wavs);
		break;
	case(FORM_AIFF):
	case(FORM_AIFC):
		rc = aiffUpdate(wavs);
		break;
	default:
		rc = RIFFWAV_E_CANT_CLOSE;
		break;
	}
	if(fclose(wavs->file))
		rc = RIFFWAV_E_CANT_CLOSE;
	wavs->file= NULL;
	return rc;
	
}


/* bend over backwards not to modify source data */
long wavWriteFloatFrames(RIFFWAV_STRUCT *wavs, const float *buf, long nFrames)
{
	long written = 0,chans,lsamp;	
	int i,j,do_reverse;
	const float *pbuf = buf;
	float fsamp,absfsamp;

	if(wavs->file==NULL || buf==NULL || nFrames < 0)
		return RIFFWAV_E_BADARG;
	chans = wavs->fmt.nChannels;
	
	switch(wavs->riff_format){
	case(RIFF_STDWAVE):		
		do_reverse = (wavs->bytemode==INORDER ? 1 : 0 );
		break;
	case(FORM_AIFF):
	case(FORM_AIFC):
		do_reverse = (wavs->bytemode==INORDER ? 0 : 1 );
		break;
	default:
		return RIFFWAV_E_UNSUPPORTED;
	}
	

	switch(wavs->samptype){
	case(SAMP_IEEE_FLOAT):		
		if(do_reverse){				
			for(i=0; i < nFrames; i++){
				for(j=0;j < chans; j++) {
					fsamp = *pbuf;
					if(wavs->clip_floats){
						fsamp = min(fsamp,1.0f);
						fsamp = max(fsamp,-1.0f);
					}
					absfsamp = (float) fabs((double)fsamp);
					if(wavs->pPeaks[j].val < absfsamp){
						wavs->pPeaks[j].pos = wavs->nFrames + i;
						wavs->pPeaks[j].val = absfsamp;
					}
					lsamp = * (long *) pbuf++;
					lsamp = REVDWBYTES(lsamp);
					if(wavDoWrite(wavs,(char *) &lsamp,sizeof(long))){
						DBGFPRINTF((stderr, "wavOpenWrite: write error\n"));
						return RIFFWAV_E_CANT_WRITE;
					}
				}
				wavs->nFrames++;								
			}
		}
		else {			
			for(i=0; i < nFrames; i++, pbuf += chans){
				for(j=0;j < chans; j++) {
					fsamp = pbuf[j];
					if(wavs->clip_floats){
						fsamp = min(fsamp,1.0f);
						fsamp = max(fsamp,-1.0f);
					}
					absfsamp = (float)fabs((double)fsamp);
					if(wavs->pPeaks[j].val < absfsamp){
						wavs->pPeaks[j].pos = wavs->nFrames + i;
						wavs->pPeaks[j].val = absfsamp;
					}					
				}								
			}
			if(wavDoWrite(wavs,(char *)buf,nFrames * chans * sizeof(float))){
				DBGFPRINTF((stderr, "wavOpenWrite: write error\n"));
				return RIFFWAV_E_CANT_WRITE;				
			}
			wavs->nFrames += nFrames;
		}
		break;
	case(SAMP_16):		
		if(do_reverse){	
			short ssamp;
			for(i=0; i < nFrames; i++){
				for(j=0;j < chans; j++) {
					fsamp = *buf++;
					/* clip now! we may have a flag to rescale first...one day */
					fsamp = min(fsamp,1.0f);
					fsamp = max(fsamp,-1.0f);
					absfsamp = (float) fabs((double)fsamp);
					if(wavs->pPeaks[j].val < absfsamp){
						wavs->pPeaks[j].pos = wavs->nFrames + i;
						wavs->pPeaks[j].val = absfsamp;
					}
					ssamp = (short) round(fsamp * 32767.0);					
					ssamp = REVWBYTES(ssamp);
					if( wavDoWrite(wavs,(char *) &ssamp,sizeof(short))){
						DBGFPRINTF((stderr, "wavOpenWrite: write error\n"));
						return RIFFWAV_E_CANT_WRITE;
					}
				}
				wavs->nFrames++;				
			}
		}
		else {
			short ssamp;
			for(i=0; i < nFrames; i++, buf += chans){
				for(j=0;j < chans; j++) {					 
					fsamp = buf[j];
					/* clip now! we may have a flag to rescale first...one day */
					fsamp = min(fsamp,1.0f);
					fsamp = max(fsamp,-1.0f);
					absfsamp = (float) fabs((double)fsamp);
					if(wavs->pPeaks[j].val < absfsamp){
						wavs->pPeaks[j].pos = wavs->nFrames + i;
						wavs->pPeaks[j].val = absfsamp;
					}
					ssamp = (short) round(fsamp * 32767.0);
					
					if(wavDoWrite(wavs,(char *) &ssamp,sizeof(short))){
						DBGFPRINTF((stderr, "wavOpenWrite: write error\n"));
						return RIFFWAV_E_CANT_WRITE;
					}
				}
				wavs->nFrames++;
				
			}			
		}
		break;
	case(SAMP_24):			 
		if(do_reverse){				
			for(i=0; i < nFrames; i++){
				for(j=0;j < chans; j++) {
					fsamp = *buf++;
					/* clip now! we may have a flag to rescale first...one day */
					fsamp = min(fsamp,1.0f);
					fsamp = max(fsamp,-1.0f);
					absfsamp = (float) fabs((double)fsamp);
					if(wavs->pPeaks[j].val < absfsamp){
						wavs->pPeaks[j].pos = wavs->nFrames + i;
						wavs->pPeaks[j].val = absfsamp;
					}
					lsamp = round(fsamp * (MAX_32BIT - 1.0f));					
					lsamp = REVDWBYTES(lsamp);					
					if( wavDoWrite(wavs,(char *) &lsamp,3)){
						DBGFPRINTF((stderr, "wavOpenWrite: write error\n"));
						return RIFFWAV_E_CANT_WRITE;
					}
				}
				wavs->nFrames++;				
			}
		}
		else {			
			for(i=0; i < nFrames; i++, buf += chans){
				for(j=0;j < chans; j++) {					 
					fsamp = buf[j];
					/* clip now! we may have a flag to rescale first...one day */
					fsamp = min(fsamp,1.0f);
					fsamp = max(fsamp,-1.0f);
					absfsamp = (float) fabs((double)fsamp);
					if(wavs->pPeaks[j].val < absfsamp){
						wavs->pPeaks[j].pos = wavs->nFrames + i;
						wavs->pPeaks[j].val = absfsamp;
					}
					lsamp = round(fsamp * (MAX_32BIT - 1.0));
					lsamp >>= 8;
					if(wavDoWrite(wavs,(char *) &lsamp,3)){
						DBGFPRINTF((stderr, "wavOpenWrite: write error\n"));
						return RIFFWAV_E_CANT_WRITE;
					}
				}
				wavs->nFrames++;				
			}			
		}
		break;
	case(SAMP_32):
		if(do_reverse){				
			for(i=0; i < nFrames; i++){
				for(j=0;j < chans; j++) {
					fsamp = *buf++;
					/* clip now! we may have a flag to rescale first...one day */
					fsamp = min(fsamp,1.0f);
					fsamp = max(fsamp,-1.0f);
					absfsamp = (float) fabs((double)fsamp);
					if(wavs->pPeaks[j].val < absfsamp){
						wavs->pPeaks[j].pos = wavs->nFrames + i;
						wavs->pPeaks[j].val = absfsamp;
					}
					lsamp = round(fsamp * (MAX_32BIT - 1.0));					
					lsamp = REVDWBYTES(lsamp);
					if( wavDoWrite(wavs,(char *) &lsamp,sizeof(long))){
						DBGFPRINTF((stderr, "wavOpenWrite: write error\n"));
						return RIFFWAV_E_CANT_WRITE;
					}
				}
				wavs->nFrames++;				
			}
		}
		else {			
			for(i=0; i < nFrames; i++, buf += chans){
				for(j=0;j < chans; j++) {					 
					fsamp = buf[j];
					/* clip now! we may have a flag to rescale first...one day */
					fsamp = min(fsamp,1.0f);
					fsamp = max(fsamp,-1.0f);
					absfsamp = (float) fabs((double)fsamp);
					if(wavs->pPeaks[j].val < absfsamp){
						wavs->pPeaks[j].pos = wavs->nFrames + i;
						wavs->pPeaks[j].val = absfsamp;
					}
					lsamp = round(fsamp * (MAX_32BIT - 1.0));
					
					if(wavDoWrite(wavs,(char *) &lsamp,sizeof(long))){
						DBGFPRINTF((stderr, "wavOpenWrite: write error\n"));
						return RIFFWAV_E_CANT_WRITE;
					}
				}
				wavs->nFrames++;				
			}			
		}
		break;
	default:
		DBGFPRINTF((stderr, "wavOpenWrite: unsupported sample format\n"));
		return RIFFWAV_E_UNSUPPORTED;		

	}
	
	return nFrames; 
		
}




long wavWriteShortFrames(RIFFWAV_STRUCT *wavs, const short *buf, long nFrames)
{
	long written = 0,chans;	 
	int i,j;
	
	if(wavs->file==NULL || buf==NULL || nFrames < 0)
		return RIFFWAV_E_BADARG;
	chans = wavs->fmt.nChannels;
		
	/* well, it can't be ~less~ efficient than converting twice! */
	if(wavs->bytemode== BYTEREV){	
		short ssamp;
		double fval;
		for(i=0; i < nFrames; i++){
			for(j=0;j < chans; j++) {
				ssamp = *buf++;
				fval = ((double) ssamp / 32768.0);
				if(wavs->pPeaks[j].val < (float)(fabs(fval))){
					wavs->pPeaks[j].pos = wavs->nFrames + i;
					wavs->pPeaks[j].val = (float)fval;
				}
								
				ssamp = REVWBYTES(ssamp);
				if(wavDoWrite(wavs,(char *) &ssamp,sizeof(short))){
					DBGFPRINTF((stderr, "wavOpenWrite: write error\n"));
					return RIFFWAV_E_CANT_WRITE;
				}
			}
			wavs->nFrames++;
			
		}
	}
	else {
		short ssamp;
		double fval;
		for(i=0; i < nFrames; i++){
			for(j=0;j < chans; j++) {
				ssamp = *buf++;
				fval = ((double) ssamp / 32768.0);
				if(wavs->pPeaks[j].val < (float)(fabs(fval))){
					wavs->pPeaks[j].pos = wavs->nFrames + i;
					wavs->pPeaks[j].val = (float)fval;
				}
				
				if(wavDoWrite(wavs,(char *) &ssamp,sizeof(short))){
					DBGFPRINTF((stderr, "wavOpenWrite: write error\n"));
					return RIFFWAV_E_CANT_WRITE;
				}
			}
			wavs->nFrames++;
			
		}			
	}
	
	return nFrames;


}

 /******** READ ***********/
static int wavReadHeader(RIFFWAV_STRUCT *wavs)
{
	long tag,version,peaktime;
	DWORD size;
	WORD cbSize;
	

	if(wavs==NULL || wavs->file == NULL)
		return RIFFWAV_E_BADARG;

	if(wavDoRead(wavs,(char *)&tag,sizeof(long))
		|| wavDoRead(wavs,(char *) &size,sizeof(long)))
		return RIFFWAV_E_CANT_READ;
	if(wavs->bytemode==BYTEREV)
		size = REVDWBYTES(size);
	if(wavs->bytemode!=INORDER)
		tag = REVDWBYTES(tag);
	if(tag != TAG('R','I','F','F'))
		return RIFFWAV_E_NOT_WAVE;
	if(size < (sizeof(WAVEFORMAT) + 3 * sizeof(WORD)))
		return RIFFWAV_E_BAD_FORMAT;

	if(wavDoRead(wavs,(char *)&tag,sizeof(long)))
		return RIFFWAV_E_CANT_READ;
	if(wavs->bytemode!=INORDER)
		tag = REVDWBYTES(tag);
	if(tag != TAG('W','A','V','E'))
		return RIFFWAV_E_NOT_WAVE;
	for(;;){
		if(wavDoRead(wavs,(char *)&tag,sizeof(long))
				|| wavDoRead(wavs,(char *) &size,sizeof(long)))
			return RIFFWAV_E_CANT_READ;
		if(wavs->bytemode==BYTEREV)
			size = REVDWBYTES(size);
		if(wavs->bytemode!=INORDER)
			tag = REVDWBYTES(tag);

		switch(tag){
		case(TAG('f','m','t',' ')):
			if( size < sizeof(WAVEFORMAT))
				return RIFFWAV_E_BAD_FORMAT;
			if(size > (sizeof(WAVEFORMAT) + sizeof(WORD)))
				return RIFFWAV_E_UNSUPPORTED;	 
			wavs->fmtoffset = ftell(wavs->file);
			if(wavDoRead(wavs,(char *)&(wavs->fmt),sizeof(WAVEFORMAT))){				
				return RIFFWAV_E_CANT_READ;
			}

			if(wavs->bytemode==BYTEREV)		
				fmtSwapBytes(wavs);
	
			/* calling function decides if format is supported*/
			if(size > sizeof(WAVEFORMAT)) {
				if(wavDoRead(wavs,(char*)&cbSize,sizeof(WORD)))
					return RIFFWAV_E_CANT_READ;
				if(cbSize != (WORD)0)
					return RIFFWAV_E_BAD_FORMAT;
			}
			/* find out what we have */
			/* ADD SUPPORT FOR MORE FORMATS HERE! */
			switch(wavs->fmt.wBitsPerSample){
			case(32):
				if(wavs->fmt.wFormatTag==WAVE_FORMAT_IEEE_FLOAT)
					wavs->samptype = SAMP_IEEE_FLOAT;
				else
					wavs->samptype = SAMP_32;
				break;
			case(24):
				wavs->samptype = SAMP_24;
				/* only support packed format for now */
				if((wavs->fmt.nBlockAlign / wavs->fmt.nChannels) != 3)
					return RIFFWAV_E_UNSUPPORTED;
				break;
			case(16):
				wavs->samptype = SAMP_16;
				break;
			/* not interested in 8bit files! */
			default:
				return RIFFWAV_E_UNSUPPORTED;
			}

			break;
		case(TAG('P','E','A','K')):
			wavs->pPeaks = (WAV_CHPEAK *) malloc(sizeof(WAV_CHPEAK) * wavs->fmt.nChannels);
			if(wavs->pPeaks==NULL){
				DBGFPRINTF((stderr, "wavOpenWrite: no memory for peak data\n"));
				return RIFFWAV_E_NOMEM;
			}
			if(size != (2 * sizeof(DWORD) + sizeof(WAV_CHPEAK) * wavs->fmt.nChannels))
				return RIFFWAV_E_BAD_FORMAT;
			if(wavDoRead(wavs,(char  *) &version,sizeof(long)))
				return RIFFWAV_E_CANT_READ;
			
			wavs->peakoffset = ftell(wavs->file);

			if(wavDoRead(wavs,(char *) &peaktime,sizeof(long)))				
				return RIFFWAV_E_CANT_READ;

			if(wavDoRead(wavs,(char *) wavs->pPeaks,sizeof(WAV_CHPEAK) * wavs->fmt.nChannels))
				return RIFFWAV_E_CANT_READ;
			if(wavs->bytemode==BYTEREV){
				DWORD *pBlock;
				int i;
				version = REVDWBYTES(version);
				if(version != 1)
					return RIFFWAV_E_UNSUPPORTED;
				wavs->peaktime = (time_t)(REVDWBYTES(peaktime));
				pBlock = (DWORD *) (wavs->pPeaks);
				for(i=0;i < wavs->fmt.nChannels * 2; i++)
					pBlock[i] = REVDWBYTES(pBlock[i]);
			}

			break;

		case(TAG('d','a','t','a')):
			wavs->dataoffset = ftell(wavs->file);
			if(wavs->fmtoffset==0)
				return RIFFWAV_E_BAD_FORMAT;
			wavs->nFrames = size / wavs->fmt.nBlockAlign;
			wavs->riff_format = RIFF_STDWAVE;
			/* get rescale factor if available */
			/* NB in correct format, val is always >= 0.0 */
			if(wavs->pPeaks && wavs->peakoffset != 0){
				float fac = 0.0f;
				int i;
				for(i=0;i < wavs->fmt.nChannels; i++)
					fac = max(fac,wavs->pPeaks[i].val);
				if(fac > 1.0f)
					wavs->rescale_fac = 1.0f / fac;
			}

			return RIFFWAV_E_NOERROR;
		default:
			/* unknown chunk - skip */
			if(fseek(wavs->file,size,SEEK_CUR))
				return RIFFWAV_E_CANT_READ;
			break;
		}
	}
}

static int aiffReadHeader(RIFFWAV_STRUCT *wavs)
{
	long tag,version,peaktime,remain,offset,blocksize;
	long have_comm =0,have_ssnd =0,have_peaks=0;	
	DWORD dwData,size;
	unsigned char ieee[10];
	WORD wData;

	if(wavs==NULL || wavs->file == NULL)
		return RIFFWAV_E_BADARG;

	if(wavDoRead(wavs,(char *)&tag,sizeof(long))
		|| wavDoRead(wavs,(char *) &remain,sizeof(long)))
		return RIFFWAV_E_CANT_READ;
	if(wavs->bytemode!=INORDER) {
		remain = REVDWBYTES(remain);	
		tag = REVDWBYTES(tag);
	}
	if(tag != TAG('F','O','R','M')){
		DBGFPRINTF((stderr, "file is not AIFF: no FORM chunk\n"));
		return RIFFWAV_E_BADARG;
	}
		
	if(wavDoRead(wavs,(char *)&tag,sizeof(long)))
		return RIFFWAV_E_CANT_READ;
	if(wavs->bytemode!=INORDER)
		tag = REVDWBYTES(tag);
	if(tag != TAG('A','I','F','F')){
		DBGFPRINTF((stderr, "file is not AIFF: no AIFF chunk\n"));
		return RIFFWAV_E_BADARG;
	}
	remain -= sizeof(long);


	while(remain > 0){
		if(wavDoRead(wavs,(char *)&tag,sizeof(long))
		|| wavDoRead(wavs,(char *) &size,sizeof(long)))
			return RIFFWAV_E_CANT_READ;
		if(wavs->bytemode!=INORDER) {
			size = REVDWBYTES(size);	
			tag = REVDWBYTES(tag);
		}
		remain -= 2 * sizeof(DWORD);
		switch(tag){
		case(TAG('C','O','M','M')):
			if(size != 18){
				DBGFPRINTF((stderr,"AIFF file has bad size for COMM chunk\n")); 
			    return RIFFWAV_E_BAD_FORMAT;
			}
			wavs->fmtoffset = ftell(wavs->file);
			if(wavDoRead(wavs,(char *)&wData,sizeof(WORD)))
				return RIFFWAV_E_CANT_READ;
			if(wavs->bytemode!=INORDER)
				wData = REVWBYTES(wData);
			wavs->fmt.nChannels = wData;
			if(wavDoRead(wavs,(char *)&dwData,sizeof(DWORD)))
				return RIFFWAV_E_CANT_READ;
			if(wavs->bytemode!=INORDER)
				dwData = REVDWBYTES(dwData);
			wavs->nFrames = dwData;
			if(wavDoRead(wavs,(char *)&wData,sizeof(WORD)))
				return RIFFWAV_E_CANT_READ;
			if(wavs->bytemode!=INORDER)
				wData = REVWBYTES(wData);
			wavs->fmt.wBitsPerSample = wData;
			if(wavDoRead(wavs,ieee,10))
				return RIFFWAV_E_CANT_READ;
			wavs->fmt.nSamplesPerSec = (DWORD)(ieee_80_to_double(ieee));
			/*we have to deduce blockalign, and hence containersize*/
			/* no support for strange wordsizes such as 20 in 24 */
			switch(wavs->fmt.wBitsPerSample){
			case(32):
				wavs->fmt.nBlockAlign = sizeof(long);
				wavs->samptype = SAMP_32;
				break;			
			case(24):
				wavs->fmt.nBlockAlign = 3;
				wavs->samptype = SAMP_24;
				break;
			case(16):
				wavs->fmt.nBlockAlign = sizeof(short);
				wavs->samptype = SAMP_16;
				break;
			//case(8):
			//	wavs->fmt.nBlockAlign = sizeof(char);
			//	wavs->samptype = SAMP_8;
			//	break;
			default:
				DBGFPRINTF((stderr, "unsupported sample format for AIFF file\n"));
				return RIFFWAV_E_UNSUPPORTED;
			}
			wavs->fmt.nBlockAlign *= wavs->fmt.nChannels;
			remain -= 18;
			have_comm = 1;
			break;
		case (TAG('P','E','A','K')):
			if(!have_comm){
				DBGFPRINTF((stderr, "AIFF file: found PEAK chunk before COMM chunk!\n"));
				return RIFFWAV_E_BAD_FORMAT;
			}
			if(size != (2 * sizeof(DWORD) + sizeof(WAV_CHPEAK) * wavs->fmt.nChannels)){
				DBGFPRINTF((stderr, "AIFF file has bad size for PEAK chunk\n"));
				return RIFFWAV_E_BAD_FORMAT;
			}
			if(wavDoRead(wavs,(char *)&version,sizeof(DWORD)))
				return RIFFWAV_E_CANT_READ;
			wavs->peakoffset = ftell(wavs->file);
			if(wavDoRead(wavs,(char *) &peaktime,sizeof(DWORD)))
				return RIFFWAV_E_CANT_READ;
			if(wavs->bytemode!=INORDER){				
				version  = REVDWBYTES(version);
				peaktime = REVDWBYTES(peaktime);
			}
			if(version != 1L){
				DBGFPRINTF((stderr, "AIFF file has unexpected version level for PEAK chunk!\n"));
				return RIFFWAV_E_UNSUPPORTED;
			}
			wavs->peaktime = (time_t) peaktime;			
			wavs->pPeaks = (WAV_CHPEAK *)malloc(sizeof(WAV_CHPEAK) * wavs->fmt.nChannels);
			if(wavs->pPeaks==NULL)
				return RIFFWAV_E_NOMEM;
			if(wavDoRead(wavs,(char *)(wavs->pPeaks),sizeof(WAV_CHPEAK) * wavs->fmt.nChannels))
				return RIFFWAV_E_CANT_READ;
			if(wavs->bytemode!=INORDER){
				DWORD *pBlock;
				int i;
				pBlock = (DWORD *) (wavs->pPeaks);
				for(i=0;i < wavs->fmt.nChannels * 2; i++)
					pBlock[i] = REVDWBYTES(pBlock[i]); 
			}
			have_peaks = 1;
			remain -= size;
			break;
		case(TAG('S','S','N','D')):
			if(wavDoRead(wavs,(char *)&offset,sizeof(long))
					|| wavDoRead(wavs,(char *) &blocksize,sizeof(long)))
			return RIFFWAV_E_CANT_READ;
			if(wavs->bytemode!=INORDER){				
				offset  = REVDWBYTES(offset);
				blocksize = REVDWBYTES(blocksize);
			}
			wavs->dataoffset = ftell(wavs->file) + offset;
			if(fseek(wavs->file,(size+1)&~1,SEEK_CUR))
				return RIFFWAV_E_CANT_SEEK;
			have_ssnd = 1;
			remain -= (size+1)&~1;
			break;
		/*HARSH! traps linux sox error, for example */
		case(0):
			DBGFPRINTF((stderr, "AIFF file has bad main chunksize\n"));
			return RIFFWAV_E_BAD_FORMAT;
		default:
			/* skip all unknown chunks */
			if(fseek(wavs->file,(size+1)&~1,SEEK_CUR))
				return RIFFWAV_E_CANT_SEEK;
			remain -= (size+1)&~1;
			break;
		}
	}
	if(!(have_ssnd && have_comm)){
		DBGFPRINTF((stderr, "AIFF file has missing chunks\n"));
		return RIFFWAV_E_BAD_FORMAT;
	}
	/* we have seeked to EOF, so rewind to start of data */
	if(fseek(wavs->file,wavs->dataoffset,SEEK_SET))
		return RIFFWAV_E_CANT_SEEK;			
	wavs->curframepos = 0;
	//wavs->filesize = wavs->nFrames * wavs->fmt.nBlockAlign;
	wavs->riff_format = FORM_AIFF;
	/* get rescale factor if available */
	/* NB in correct format, val is always >= 0.0 */
	if(wavs->pPeaks && wavs->peakoffset != 0){
		float fac = 0.0f;
		int i;
		for(i=0;i < wavs->fmt.nChannels; i++)
			fac = max(fac,wavs->pPeaks[i].val);
		if(fac > 1.0f)
			wavs->rescale_fac = 1.0f / fac;
	}
	return RIFFWAV_E_NOERROR;
}

static int aifcReadHeader(RIFFWAV_STRUCT *wavs)
{
	long tag,version,peaktime,remain,offset,blocksize;
	long have_comm =0,have_ssnd =0,have_peaks=0,have_fver = 0;	
	DWORD dwData,size,aifcver,ID_compression;
	unsigned char ieee[10];
	WORD wData;

	if(wavs==NULL || wavs->file == NULL)
		return RIFFWAV_E_BADARG;

	if(wavDoRead(wavs,(char *)&tag,sizeof(long))
		|| wavDoRead(wavs,(char *) &remain,sizeof(long)))
		return RIFFWAV_E_CANT_READ;
	if(wavs->bytemode!=INORDER) {
		remain = REVDWBYTES(remain);	
		tag = REVDWBYTES(tag);
	}
	if(tag != TAG('F','O','R','M')){
		DBGFPRINTF((stderr, "file is not AIFC: no FORM chunk\n"));
		return RIFFWAV_E_BADARG;
	}
	
	if(wavDoRead(wavs,(char *)&tag,sizeof(long)))
		return RIFFWAV_E_CANT_READ;
	if(wavs->bytemode!=INORDER)
		tag = REVDWBYTES(tag);
	if(tag != TAG('A','I','F','C')){
		DBGFPRINTF((stderr, "file is not AIFC: no AIFC chunk\n"));
		return RIFFWAV_E_BADARG;
	}
	remain -= sizeof(long);


	while(remain > 0){
		if(wavDoRead(wavs,(char *)&tag,sizeof(long))
		|| wavDoRead(wavs,(char *) &size,sizeof(long)))
			return RIFFWAV_E_CANT_READ;
		if(wavs->bytemode!=INORDER) {
			size = REVDWBYTES(size);	
			tag = REVDWBYTES(tag);
		}
		remain -= 2 * sizeof(DWORD);
		switch(tag){
		case(TAG('F','V','E','R')):
			if(size != sizeof(DWORD)){
				DBGFPRINTF((stderr, "AIFC file has bad size for FVER chunk\n"));
				return RIFFWAV_E_BAD_FORMAT;
			}
			if(wavDoRead(wavs,(char *) &aifcver,sizeof(DWORD)))
				return RIFFWAV_E_CANT_READ;
			if(wavs->bytemode!=INORDER) 
				aifcver = REVDWBYTES(aifcver);
			remain-= sizeof(DWORD);
			if(aifcver != AIFC_VERSION_1)
				return RIFFWAV_E_UNSUPPORTED;
			have_fver = 1;
			break;
		case(TAG('C','O','M','M')):
			if(size < 22) {
			   DBGFPRINTF((stderr, "AIFC file has bad size for COMM chunk\n"));
			   return RIFFWAV_E_BAD_FORMAT;
			}
			wavs->fmtoffset = ftell(wavs->file);
			if(wavDoRead(wavs,(char *)&wData,sizeof(WORD)))
				return RIFFWAV_E_CANT_READ;
			if(wavs->bytemode!=INORDER)
				wData = REVWBYTES(wData);
			wavs->fmt.nChannels = wData;
			if(wavDoRead(wavs,(char *)&dwData,sizeof(DWORD)))
				return RIFFWAV_E_CANT_READ;
			if(wavs->bytemode!=INORDER)
				dwData = REVDWBYTES(dwData);
			wavs->nFrames = dwData;
			if(wavDoRead(wavs,(char *)&wData,sizeof(WORD)))
				return RIFFWAV_E_CANT_READ;
			if(wavs->bytemode!=INORDER)
				wData = REVWBYTES(wData);
			wavs->fmt.wBitsPerSample = wData;
			if(wavDoRead(wavs,ieee,10))
				return RIFFWAV_E_CANT_READ;
			wavs->fmt.nSamplesPerSec = (DWORD)(ieee_80_to_double(ieee));
			/*we have to deduce blockalign, and hence containersize*/
			/* no support for strange wordsizes such as 20 in 24 */
			switch(wavs->fmt.wBitsPerSample){
			case(32):
				wavs->fmt.nBlockAlign = sizeof(long);
				wavs->samptype = SAMP_32;
				break;			
			case(24):
				wavs->fmt.nBlockAlign = 3;
				wavs->samptype = SAMP_24;
				break;
			case(16):
				wavs->fmt.nBlockAlign = sizeof(short);
				wavs->samptype = SAMP_16;
				break;
			//case(8):
			//	wavs->fmt.nBlockAlign = sizeof(char);
			//	wavs->samptype = SAMP_8;
			//	break;
			default:
				DBGFPRINTF((stderr, "unsupported sample format for AIFC file\n"));
				return RIFFWAV_E_UNSUPPORTED;
			}
			wavs->fmt.nBlockAlign *= wavs->fmt.nChannels;
			if(wavDoRead(wavs,(char *)&ID_compression,sizeof(DWORD))){
				return RIFFWAV_E_CANT_READ;	
			}
			if(wavs->bytemode!=INORDER)
				ID_compression = REVDWBYTES(ID_compression);
			if(!(    (ID_compression == TAG('N','O','N','E'))
				  || (ID_compression == TAG('F','L','3','2'))	//Csound,SGI,Soundhack,CDP
				  || (ID_compression == TAG('f','l','3','2')))	//Apple,of course
						  ){
				DBGFPRINTF((stderr, "AIFC file: unsupported compression format\n"));
				return RIFFWAV_E_UNSUPPORTED;
			}
			/*set stype info */
			if((ID_compression== TAG('F','L','3','2'))		  
				|| ID_compression == TAG('f','l','3','2')){
					if(wavs->fmt.wBitsPerSample != 32){						
						DBGFPRINTF((stderr, "AIFC file: samples not 32bit in floats file\n"));
						return RIFFWAV_E_BAD_FORMAT;
					}
					else {
						wavs->fmt.wFormatTag = WAVE_FORMAT_IEEE_FLOAT;			
						wavs->samptype = SAMP_IEEE_FLOAT;
					}
			}
			/* yes, lazy! skip past pascal string*/
			if(fseek(wavs->file,((size-22)+1)&~1,SEEK_CUR))	/*written for documentation, not terseness!*/
				return RIFFWAV_E_CANT_SEEK;
			remain -= (size+1)&~1;
			have_comm = 1;
			break;
		case (TAG('P','E','A','K')):
			if(!have_comm){
				DBGFPRINTF((stderr, "\nAIFC file: found PEAK chunk before COMM chunk!\n"));
				return RIFFWAV_E_BAD_FORMAT;
			}
			if(size != (2 * sizeof(DWORD) + sizeof(WAV_CHPEAK) * wavs->fmt.nChannels)){
				DBGFPRINTF((stderr, "\nAIFC file has bad size for PEAK chunk\n"));
				return RIFFWAV_E_BAD_FORMAT;
			}
			if(wavDoRead(wavs,(char *)&version,sizeof(DWORD)))
				return RIFFWAV_E_CANT_READ;
			wavs->peakoffset = ftell(wavs->file);
			if(wavDoRead(wavs,(char *) &peaktime,sizeof(DWORD)))
				return RIFFWAV_E_CANT_READ;
			if(wavs->bytemode!=INORDER){				
				version  = REVDWBYTES(version);
				peaktime = REVDWBYTES(peaktime);
			}
			if(version != 1L) {
				DBGFPRINTF((stderr, "AIFC has unexpected version level for PEAK chunk!\n"));
				return RIFFWAV_E_UNSUPPORTED;
			}
			wavs->peaktime = (time_t) peaktime;			
			wavs->pPeaks = (WAV_CHPEAK *)malloc(sizeof(WAV_CHPEAK) * wavs->fmt.nChannels);
			if(wavs->pPeaks==NULL)
				return RIFFWAV_E_NOMEM;
			if(wavDoRead(wavs,(char *)(wavs->pPeaks),sizeof(WAV_CHPEAK) * wavs->fmt.nChannels))
				return RIFFWAV_E_CANT_READ;
			if(wavs->bytemode!=INORDER){
				DWORD *pBlock;
				int i;
				pBlock = (DWORD *) (wavs->pPeaks);
				for(i=0;i < wavs->fmt.nChannels * 2; i++)
					pBlock[i] = REVDWBYTES(pBlock[i]); 
			}
			have_peaks = 1;
			remain -= size;
			break;
		case(TAG('S','S','N','D')):
			if(wavDoRead(wavs,(char *)&offset,sizeof(long))
					|| wavDoRead(wavs,(char *) &blocksize,sizeof(long)))
			return RIFFWAV_E_CANT_READ;
			if(wavs->bytemode!=INORDER){				
				offset  = REVDWBYTES(offset);
				blocksize = REVDWBYTES(blocksize);
			}
			wavs->dataoffset = ftell(wavs->file) + offset;
			if(fseek(wavs->file,(size+1)&~1,SEEK_CUR))
				return RIFFWAV_E_CANT_SEEK;
			have_ssnd = 1;
			remain -= (size+1)&~1;
			break;
		/* HARSH! as above */
		case(0):
			DBGFPRINTF((stderr, "AIFC file has bad main chunksize\n"));
			return RIFFWAV_E_BAD_FORMAT;
		default:
			/* skip all unknown chunks */
			if(fseek(wavs->file,(size+1)&~1,SEEK_CUR))
				return RIFFWAV_E_CANT_SEEK;
			remain -= (size+1)&~1;
			break;
		}
	}
	if(!(have_ssnd && have_comm && have_fver)){
		DBGFPRINTF((stderr, "AIFC file has bad format\n"));
		return RIFFWAV_E_BAD_FORMAT;
	}
	/* we have seeked to EOF, so rewind to start of data */
	if(fseek(wavs->file,wavs->dataoffset,SEEK_SET))
		return RIFFWAV_E_CANT_SEEK;			
	wavs->curframepos = 0;
	//wavs->filesize = wavs->nFrames * wavs->fmt.nBlockAlign;
	wavs->riff_format = FORM_AIFC;
	/* get rescale factor if available */
	/* NB in correct format, val is always >= 0.0 */
	if(wavs->pPeaks && wavs->peakoffset != 0){
		float fac = 0.0f;
		int i;
		for(i=0;i < wavs->fmt.nChannels; i++)
			fac = max(fac,wavs->pPeaks[i].val);
		if(fac > 1.0f)
			wavs->rescale_fac = 1.0f / fac;
	}
	return RIFFWAV_E_NOERROR;
}


int wavOpenRead(RIFFWAV_STRUCT *wavs, const char *path, int rescale)
{
	sf_format fmt;
	
	if(wavs == NULL || path==NULL)
		return RIFFWAV_E_BADARG;
	if(wavs->file)
		return RIFFWAV_E_CANT_OPEN;
	wavs->rescale = rescale;	
	wavs->bytemode = HostByteMode();
	fmt = get_format_ext(path);
	if(!(fmt==RIFF_STDWAVE || fmt==FORM_AIFF || fmt==FORM_AIFC))
		return RIFFWAV_E_BADARG;
	
	if((wavs->file = fopen(path,"rb"))  == NULL) {
		DBGFPRINTF((stderr, "wavOpenWrite: cannot create '%s'\n", path));
        return RIFFWAV_E_CANT_OPEN;
	}
	wavs->filename = (char *) malloc(strlen(path)+1);
	if(wavs->filename==NULL) {
		DBGFPRINTF((stderr, "wavOpenRead: no memory for filename\n"));
		return RIFFWAV_E_NOMEM;
	}
    strcpy(wavs->filename, path);
    wavs->isRead = wavs->isSeekable = 1;	
	wavs->filesize = 0;
	wavs->nFrames = 0;
	/* no need to calc header sizes */
	switch(fmt){
	case(RIFF_STDWAVE):
	case(RIFF_WAVE_EX):
		return wavReadHeader(wavs);
	case(FORM_AIFF):
		return aiffReadHeader(wavs);
	case(FORM_AIFC):
		return aifcReadHeader(wavs);
	default:
		DBGFPRINTF((stderr, "wavOpenRead: unsupported file format\n"));
		return RIFFWAV_E_UNSUPPORTED;
	}
}



long wavReadFloatFrames(RIFFWAV_STRUCT *wavs, float *buf, long nFrames)
{
	long chans,framesread;
	long blocksize,lsamp;
	int i,do_reverse;
	short ssamp;
	float *pbuf = buf;
	float fsamp;
	
	
	if(wavs->file==NULL || buf==NULL || nFrames < 0)
		return RIFFWAV_E_BADARG;
	chans = wavs->fmt.nChannels;
	framesread = MIN(wavs->nFrames - wavs->curframepos,nFrames);
	blocksize =  framesread * chans;
	if(framesread==0)
		return framesread;

	switch(wavs->riff_format){
	case(RIFF_STDWAVE):		
		do_reverse = (wavs->bytemode==INORDER ? 1 : 0 );
		break;
	case(FORM_AIFF):
	case(FORM_AIFC):
		do_reverse = (wavs->bytemode==INORDER ? 0 : 1 );
		break;
	default:
		return RIFFWAV_E_UNSUPPORTED;
	}

	switch(wavs->samptype){
	case(SAMP_IEEE_FLOAT):
		
		if(do_reverse){
			for(i=0;i < blocksize;i ++){
				if(wavDoRead(wavs,(char *)&lsamp,sizeof(long)))
					return RIFFWAV_E_CANT_READ;
				lsamp = REVDWBYTES(lsamp);
				fsamp = * (float *)&lsamp;
				if(wavs->rescale)
					fsamp *= wavs->rescale_fac;
				*pbuf++ = fsamp;
			}			
		}
		else{
			if(wavDoRead(wavs,(char *) buf,blocksize * sizeof(float)))
				return RIFFWAV_E_CANT_READ;
			if(wavs->rescale){
				pbuf = buf;
				for(i=0;i < blocksize; i++)
					*pbuf++  *= wavs->rescale_fac;
			}
		}
		break;
	case(SAMP_16):
		if(do_reverse){
			for(i = 0; i < blocksize; i++){
				if(wavDoRead(wavs,(char *)&ssamp,sizeof(short)))
					return RIFFWAV_E_CANT_READ;
				ssamp = REVWBYTES(ssamp);
				fsamp = (float)((double) ssamp / MAX_16BIT);
				*pbuf++ = fsamp;
			}
		}
		else{
			for(i = 0; i < blocksize; i++){
				if(wavDoRead(wavs,(char *)&ssamp,sizeof(short)))
					return RIFFWAV_E_CANT_READ;				
				fsamp = (float)((double) ssamp / MAX_16BIT);
				*pbuf++ = fsamp;
			}
		}
		break;
	case(SAMP_24):
		if(do_reverse){
			for(i=0;i < blocksize;i++){
				if(wavDoRead(wavs,(char *)&lsamp,3))
					return RIFFWAV_E_CANT_READ;
				lsamp = REVDWBYTES(lsamp);
				lsamp <<= 8;
				fsamp = (float)((double)(lsamp) / MAX_32BIT);
				*pbuf++ = fsamp;
			}
		}
		else{
			for(i=0;i < blocksize;i++){
				lsamp = 0L;
				if(wavDoRead(wavs,(char *)&lsamp,3))
					return RIFFWAV_E_CANT_READ;
				lsamp <<= 8;
				fsamp = (float)((double)(lsamp) / MAX_32BIT);
				*pbuf++ = fsamp;
			}
		}
		break;
	case(SAMP_32):
		if(do_reverse){
			for(i=0;i < blocksize;i++){
				if(wavDoRead(wavs,(char *)&lsamp,sizeof(long)))
					return RIFFWAV_E_CANT_READ;
				lsamp = REVDWBYTES(lsamp);				
				fsamp = (float)((double)(lsamp) / MAX_32BIT);
				*pbuf++ = fsamp;
			}
		}
		else{
			for(i=0;i < blocksize;i++){
				lsamp = 0L;
				if(wavDoRead(wavs,(char *)&lsamp,sizeof(long)))
					return RIFFWAV_E_CANT_READ;				
				fsamp = (float)((double)(lsamp) / MAX_32BIT);
				*pbuf++ = fsamp;
			}
		}
		break;
	default:
		DBGFPRINTF((stderr, "wavOpenRead: unsupported sample format\n"));
		return RIFFWAV_E_UNSUPPORTED;
	}	
	wavs->curframepos += framesread;
	return framesread;
}


/* returns multi-channel (frame)  position */
long wavTell(RIFFWAV_STRUCT *wavs)
{
	long pos;
	if(wavs==NULL || wavs->file==NULL)
		return RIFFWAV_E_BADARG;
	pos = ftell(wavs->file);
	pos-= wavs->dataoffset;
	pos /= wavs->fmt.nBlockAlign;
	return MAX(pos,0);			 /* should never be negative...*/

}

/* NB this does not handle seeks beyond current end of file */
long wavSeek(RIFFWAV_STRUCT *wavs,long to)
{
	DWORD bytepos,at;
	if(wavs==NULL || wavs->file==NULL || wavs->dataoffset==0)
		return RIFFWAV_E_BADARG;
	bytepos = to *  wavs->fmt.nBlockAlign;
	bytepos += wavs->dataoffset;
	if(fseek(wavs->file,bytepos,SEEK_SET))
		return RIFFWAV_E_CANT_SEEK;
	at = ftell(wavs->file);
	if(at >= wavs->dataoffset) {
		wavs->curframepos = (at - wavs->dataoffset) / wavs->fmt.nBlockAlign;
		return wavs->curframepos;
	}
	else return RIFFWAV_E_CANT_SEEK;

}



sf_format get_format_ext(const char *path)
{
	char *lastdot;
	if(path==NULL || (strlen(path) < 4))
		return FMT_UNKNOWN;				//better than no err-check at all!
	lastdot = strrchr(path,'.');
	if(lastdot==NULL)
		return FMT_UNKNOWN;

	if(stricmp(lastdot,".wav")==0)
		 return RIFF_STDWAVE;
	else if((stricmp(lastdot,".aif")==0) || stricmp(lastdot,".aiff")==0)
		 return FORM_AIFF;
	else if((stricmp(lastdot,".afc")==0) || stricmp(lastdot,".aifc")==0)
			return FORM_AIFC;
	else
		return FMT_UNKNOWN;

}

int wavSetFormat(RIFFWAV_STRUCT *wavs,sf_stype stype,long channels, long srate)
{
	if(wavs==NULL)
		return RIFFWAV_E_BADARG;
	if(wavs->file)		/*already opened */
		return RIFFWAV_E_BADARG;

	wavs->samptype					= stype;
	wavs->fmt.nSamplesPerSec		= (DWORD) srate;
	wavs->fmt.nChannels				= (WORD) channels;
	wavs->fmt.wFormatTag			= WAVE_FORMAT_PCM;
	switch(stype){
	case(SAMP_16):
		wavs->fmt.wBitsPerSample	= (WORD)(sizeof(short) * BITS_PER_BYTE);
		wavs->fmt.nBlockAlign		= (WORD)(sizeof(short) * channels);
		break;
	case(SAMP_24):
		wavs->fmt.wBitsPerSample	= (WORD)(3 * BITS_PER_BYTE);
		wavs->fmt.nBlockAlign		= (WORD)(3 * channels);
		break;
	case(SAMP_32):
		wavs->fmt.wBitsPerSample	= (WORD)(sizeof(long) * BITS_PER_BYTE);
		wavs->fmt.nBlockAlign		= (WORD)(sizeof(long) * channels);
		break;
	case(SAMP_IEEE_FLOAT):
		wavs->fmt.wBitsPerSample	= (WORD)(sizeof(float) * BITS_PER_BYTE);
		wavs->fmt.nBlockAlign		= (WORD)(sizeof(float) * channels);
		wavs->fmt.wFormatTag		= WAVE_FORMAT_IEEE_FLOAT;
		break;
	default:
		return RIFFWAV_E_UNSUPPORTED;
	}
	wavs->fmt.nAvgBytesPerSec		= (DWORD)(wavs->fmt.nBlockAlign * wavs->fmt.nSamplesPerSec);
	return RIFFWAV_E_NOERROR;
}
