	#ifndef __RIFFWAV_H_INCLUDED
#define __RIFFWAV_H_INCLUDED
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dpwelib.h"
#include "byteswap.h"



#ifndef WORD
#define WORD unsigned short
#endif

#ifndef DWORD
#define DWORD unsigned long
#endif

/* the file sample formats we could support */

typedef enum {
	SAMP_UNKNOWN	=	0,
	SAMP_8,
	SAMP_16,
	SAMP_24,
	SAMP_2432,		/*NB AIFF can't do this one */
	SAMP_32,
	SAMP_IEEE_FLOAT
} sf_stype;


typedef enum {
	FMT_UNKNOWN = 0,		/* e.g if no extension given */
	RIFF_STDWAVE,
	RIFF_WAVE_EX,
	FORM_AIFF,
	FORM_AIFC
} sf_format;

/* error codes - no positive ones!*/
enum {
	RIFFWAV_E_NOERROR		= 0,
	RIFFWAV_E_CANT_OPEN		= -1,
	RIFFWAV_E_CANT_CLOSE	= -2,
	RIFFWAV_E_CANT_WRITE	= -3,
	RIFFWAV_E_CANT_READ		= -4,
	RIFFWAV_E_NOT_WAVE		= -5,
	RIFFWAV_E_BAD_TYPE		= -6,
	RIFFWAV_E_BAD_FORMAT	= -7,
	RIFFWAV_E_UNSUPPORTED	= -8,
	RIFFWAV_E_NOMEM			= -9,
	RIFFWAV_E_BADARG		= -10,
	RIFFWAV_E_CANT_SEEK		= -11
};

/* the standard structures, from Windows mmsystem.h, etc*/

typedef struct riffwav_waveformat {
	WORD  wFormatTag; 
    WORD  nChannels; 
    DWORD nSamplesPerSec; 
    DWORD nAvgBytesPerSec; 
    WORD  nBlockAlign; 
    WORD  wBitsPerSample; 

} WAVEFORMAT;
/* may need this too */
typedef struct	riffwav_waveformatex {
	WAVEFORMAT fmt;	 
    WORD  cbSize;

} WAVEFORMATEX;

#define WAVE_FORMAT_PCM			(0x0001)
#define WAVE_FORMAT_IEEE_FLOAT	(0x0003)
/*I've, started, so I'll finish some day */
#define WAVE_FORMAT_EXTENSIBLE	(0xfffe)


/* support for the PEAK chunk */
typedef struct wav_chpeak {	
	float val;
	DWORD pos;

} WAV_CHPEAK;



typedef struct riffwav_struct {
	FILE			*file;
	char			*filename;
	DWORD			filesize;		/* bytes, incl header */
	DWORD			curframepos;	/* for read operations */
	DWORD			nFrames;		/* multi-channel sample frames */
	int				isRead;
	int				clip_floats;
	int				rescale;
	float			rescale_fac;
	sf_format		riff_format;
	int				isSeekable;
	int				bytemode;		/* how to swap bytes (e.g. on DEC) */
	sf_stype		samptype;		/* = nBlockAlign / nChannels */
	DWORD			dataoffset;		/* = sizeof(header) */
	DWORD			fmtoffset;
	DWORD			peakoffset;
	WAVEFORMAT		fmt;
	WAV_CHPEAK		*pPeaks;
	time_t			peaktime;
} RIFFWAV_STRUCT;

/*RWD in time, all this will be replaced by generic snd functions! */

RIFFWAV_STRUCT *wavNew(void);
void wavFree(RIFFWAV_STRUCT *wavs);

int wavOpenWrite(RIFFWAV_STRUCT *wavs, const char *path,sf_stype stype,int clip_floats);
/* I want an eror return */
int wavClose(RIFFWAV_STRUCT *wavs);
/* RWD not interested in void pointers....*/
/* doesn't this make sfys look good! */
long wavWriteFloatFrames(RIFFWAV_STRUCT *wavs, const float *buf, long nFrames);
long wavWriteShortFrames(RIFFWAV_STRUCT *wavs, const short *buf, long nFrames);
long wavTell(RIFFWAV_STRUCT *wavs);
long wavSeek(RIFFWAV_STRUCT *wavs,long to);
int wavOpenRead(RIFFWAV_STRUCT *wavs, const char *path,int rescale);
long wavReadFloatFrames(RIFFWAV_STRUCT *wavs, float *buf, long nFrames);
int wavSetFormat(RIFFWAV_STRUCT *wavs,sf_stype stype,long channels, long srate); 
/* general usage */
sf_format get_format_ext(const char *path);

#endif